#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>

#include "mtest.h"

#define INDEX(r) ((r) % alloc_number)
#define SIZE(r)  (((r) / alloc_number) & (alloc_maxsize - 1))

#ifndef CHECK_CONTENTS
#define CHECK_CONTENTS 0
#endif

static void **allocs;
static int *allocs_size;

static void test_check(int index, int size)
{
	int i;
	unsigned char b = index;
	unsigned char *p;

	if (size == -1) {
		/* test on free, data must be the same as above */
		size = allocs_size[index];
		p = allocs[index];
		for (i = 0; i < size; i++, b++) {
			if (p[i] != b) {
				printf("ERROR: index %i offset %i\n",
				       index, i);
			}
		}
		return;
	}
	/* alloc: fill the area */
	p = allocs[index];
	for (i = 0; i < size; i++, b++)
		p[i] = b;
}

void testloop(int alloc_number, int alloc_maxsize, int alloc_loops,
	      int alloc_verbose)
{
	struct timeval tv1, tv2;
	int i, count, r, size, nalloc, nfree, nfail;
	void *min, *max;

	min = max = sbrk_init(alloc_maxsize * alloc_number);
	if (min == (void*)-1)
		exit(2);
	allocs = malloc(alloc_number * sizeof(*allocs));
	if (!allocs) exit(1);
	/* mlockall may have failed, touch all memory to be sure */
	memset(allocs, 0, alloc_number * sizeof(*allocs));

	if (CHECK_CONTENTS) {
		/* We must remember the size as well */
		allocs_size = malloc(alloc_number * sizeof(int));
	}

	/*
	 * Loop: try to randomize allocs and frees. Size
	 * is randomized and index is randomized as well.
	 */
	setvbuf(stdout, NULL, _IONBF, 0);
	nalloc = 0; nfree = 0; nfail = 0;
	gettimeofday(&tv1, NULL);
	for (count = 0; count < alloc_loops; count++) {
		r = rand();
		i = INDEX(r);
		size = SIZE(r);
		if (allocs[i]) {
			nfree++;
			if (alloc_verbose)
				printf("index %4i freeing %08x\n", i,
				       allocs[i]-min);
			dlfree(allocs[i]);
			allocs[i] = 0;
		} else {
			allocs[i] = dlmalloc(size);
			if (!allocs[i]) {
				nfail++;
				if (CHECK_CONTENTS)
					test_check(i, -1);
			} else {
				nalloc++;
				if (allocs[i] + size > max)
					max = allocs[i] + size;
			}
			if (CHECK_CONTENTS)
				test_check(i, size);
			if (alloc_verbose)
				printf("index %4i allocated %08x for 0x%04x\n",
				       i, allocs[i]-min, size);
		}
		if (alloc_verbose > 1)
			rbm_show(stdout, min);
	}
	gettimeofday(&tv2, NULL);

	if (alloc_verbose) {
		printf("\n\n**** final status:\n");
		rbm_show(stdout, min);
	}

	if (CHECK_CONTENTS)
		for (i = 0; i < alloc_number; i++)
			if (allocs[i])
				test_check(i, -1);

	printf("%i random ops (%i alloc, %i fail, %i free)\n", count,
	       nalloc, nfail, nfree);
	printf("T %9li    S %9li\n", tv2.tv_usec - tv1.tv_usec +
	       1000*1000*(tv2.tv_sec - tv1.tv_sec),
	       max - min);
}
